#!/usr/bin/python3 -t
# -*- coding: utf-8 -*-

import boto3
import logging
import subprocess
import json
import uuid
import argparse
import os
import ipaddress

PSKD='DEGUPSK'
JOIN_TIMEOUT='30'
CONFIG_DIR="/etc/coap-mqtt/"
ASSET_DIR="/tmp/asset/"
KEY_EXT="key.pem"
CSR_EXT="csr.pem"
CERT_EXT="crt.pem"
A71CH_CMD="a71chConfig_i2c_imx"
A71CH_RESET=A71CH_CMD+" debug reset"
A71CH_SET_PAIR=A71CH_CMD+" set pair -x 0 -k "
A71CH_REFPEM=A71CH_CMD+" refpem -c 10 -x 0 -r "

class DeguMgr:
    def __init__(self):
        f = open(CONFIG_DIR + 'mqttinfo.json')
        p = json.load(f)
        self.iot = boto3.client('iot',
                                aws_access_key_id=p['accesskeyid'],
                                aws_secret_access_key=p['secretaccesskey'],
                                region_name=p['region']
        )
        self.sts = boto3.client('sts',
                                aws_access_key_id=p['accesskeyid'],
                                aws_secret_access_key=p['secretaccesskey'],
                                region_name=p['region']
        )
        self.accesskeyid = p['accesskeyid']
        self.region = p['region']
        self.gw_preset = p['gw']['preset']
        self.gw_mqtt = p['gw']['mqtt']

    def exists(self, name):
        list_of_things = self.iot.list_things()
        for thing in list_of_things['things']:
            if thing['thingName'] == name:
                return True
        while 'nextToken' in list_of_things:
            t = list_of_things['nextToken']
            list_of_things = self.iot.list_things(nextToken=t)
            for thing in list_of_things['things']:
                if thing['thingName'] == name:
                    return True
        return False

    def prepare_thing_types(self):
        self.iot.create_thing_type(thingTypeName='Degu')
        self.iot.create_thing_type(thingTypeName='Armadillo')

        return True

    def gw_ca_key_name(self):
        return self.gw_preset["ca"]["key"]

    def gw_ca_cert_name(self):
        return self.gw_preset["ca"]["cert"]

    def gw_device_key_name(self):
        if self.hasDeviceKey() and self.hasDeviceCert():
            return self.gw_preset["device"]["key"]
        return CONFIG_DIR + "degugw_device" + "." + KEY_EXT

    def gw_device_csr_name(self):
        return CONFIG_DIR + "degugw_device" + "." + CSR_EXT

    def gw_device_cert_name(self):
        if self.hasDeviceKey() and self.hasDeviceCert():
            return self.gw_preset["device"]["cert"]
        return CONFIG_DIR + "degugw_device" + "." + CERT_EXT

    def gw_device_ref_name(self):
        return self.gw_mqtt["key"]

    def gw_device_and_ca_name(self):
        return self.gw_mqtt["cert"]

    def create_device_cert_for_gw(self,mac):
        out = subprocess.getoutput("openssl ecparam -genkey -name prime256v1 -out " + self.gw_device_key_name())
        out = subprocess.getoutput("openssl req -new -key " + self.gw_device_key_name() + " -out " + self.gw_device_csr_name() + " -subj \"/CN=" + mac + "\"")
        out = subprocess.getoutput("openssl x509 -req -in " + self.gw_device_csr_name() + " -CA " + self.gw_ca_cert_name() + " -CAkey " + self.gw_ca_key_name() + " -CAcreateserial -out " + self.gw_device_cert_name() + " -days 3650 -sha256")

    def register_device_cert_for_gw(self):
        ca_certificate_filename = self.gw_ca_cert_name()
        device_certificate_filename = self.gw_device_cert_name()
        ca_certificate_file = open(ca_certificate_filename, 'r')
        device_certificate_file = open(device_certificate_filename, 'r')
        ca_certificate_text = ca_certificate_file.read()
        device_certificate_text = device_certificate_file.read()

        result = self.iot.register_certificate(
                    certificatePem=device_certificate_text,
                    caCertificatePem=ca_certificate_text,
                    status='ACTIVE')

        arn = result['certificateArn']
        self.arn = arn
        print(result['certificateId'])

        return True

    def create_attach_assets_degu(self, mac, path):
        result = self.create_keys_and_certificate()
        cert_id = result['certificateId']

        cert_pem = result['certificatePem']
        cert_pem_filename = path + '/certificate.pem.crt'
        cert_pem_file = open(cert_pem_filename, 'w')
        cert_pem_file.write(cert_pem)
        cert_pem_file_path = os.path.abspath(cert_pem_filename)
        os.chmod(cert_pem_file_path, 0o600)
        cert_pem_file.close()

        private_key_pem = result['keyPair']['PrivateKey']
        private_key_pem_filename = path + '/private.pem.key'
        private_key_pem_file = open(private_key_pem_filename, 'w')
        private_key_pem_file.write(private_key_pem)
        private_key_pem_file_path = os.path.abspath(private_key_pem_filename)
        os.chmod(private_key_pem_file_path, 0o600)
        private_key_pem_file.close()

        public_key_pem = result['keyPair']['PublicKey']
        public_key_pem_filename = path + '/public.pem.key'
        public_key_pem_file = open(public_key_pem_filename, 'w')
        public_key_pem_file.write(public_key_pem)
        public_key_pem_file_path = os.path.abspath(public_key_pem_filename)
        os.chmod(public_key_pem_file_path, 0o600)
        public_key_pem_file.close()

        self.create_attach_policy("degu")

        return True

    def create_attach_assets_degugw(self, mac):
        if not self.hasDeviceKey() or not self.hasDeviceCert():
            self.create_device_cert_for_gw(mac)
        self.register_device_cert_for_gw()
        self.create_attach_policy("DeguGW")

        return True

    def hasDeviceKey(self):
        if not self.gw_preset["device"]["key"]:
            return False
        if not os.path.isfile(self.gw_preset["device"]["key"]):
            return False
        return True

    def hasDeviceCert(self):
        if not self.gw_preset["device"]["cert"]:
            return False
        if not os.path.isfile(self.gw_preset["device"]["cert"]):
            return False
        return True

    def hasCaKey(self):
        if not self.gw_preset["ca"]["key"]:
            return False
        if not os.path.isfile(self.gw_preset["ca"]["key"]):
            return False
        return True

    def hasCaCert(self):
        if not self.gw_preset["ca"]["cert"]:
            return False
        if not os.path.isfile(self.gw_preset["ca"]["cert"]):
            return False
        return True

    def set_assets_to_a71ch(self):
        out = subprocess.getoutput(A71CH_SET_PAIR + self.gw_device_key_name())
        out = subprocess.getoutput(A71CH_REFPEM + self.gw_device_ref_name())
        out = subprocess.getoutput("cat " + self.gw_device_cert_name() + " " + self.gw_ca_cert_name() + " > " + self.gw_device_and_ca_name())

    def create_attach_policy(self, policy):
        private_key_pem_filename = '/usr/share/degu-manager/policy_document.templ'
        private_key_pem_file = open(private_key_pem_filename, 'r')
        id_info = self.sts.get_caller_identity()
        policy_document_text = private_key_pem_file.read().replace('<region>', self.region).replace('<accountid>', id_info['Account'])
        try:
            self.iot.create_policy(policyName=policy, policyDocument=policy_document_text)
        except:
            pass
        result = self.iot.attach_principal_policy(policyName=policy, principal=self.arn)

        return True

    def attach_gw(self, mac, name):
        self.create_attach_assets_degugw(name)
        if not self.hasCaCert():
            print("no CA")
            return False
        self.attach_thing("Armadillo-" + mac)
        self.set_assets_to_a71ch()

    def register_gw(self, mac):
        self.prepare_thing_types()
        name = 'Armadillo-'+mac
        if self.exists(name):
            print("gateway already exists")
            result, status = self.validate_certificate(name)
            if status == "None":
                self.attach_gw(mac, name)
            return False
        resp = self.iot.create_thing(thingName=name, thingTypeName='Armadillo')
        self.attach_gw(mac, name)

        return True

    def register_degu(self, mac):
        self.prepare_thing_types()
        try:
            resp = self.iot.create_thing(thingName=mac,thingTypeName='Degu')
        except Exception as e:
            if e.response['Error']['Code'] == 'ResourceAlreadyExistsException':
                print("User already exists")
            else:
                print("Unexpected error: %s" % e)

        os.makedirs(ASSET_DIR + mac, exist_ok=True)
        self.create_attach_assets_degu(mac, ASSET_DIR + mac)
        self.attach_thing(mac)

        return True

    def unregister_gw(self, mac):
        try:
            resp = self.iot.delete_thing(thingName='Armadillo-'+mac)
        except Exception as e:
            if e.response['Error']['Code'] == 'ResourceAlreadyExistsException':
                print("User already exists")
            else:
                print("Unexpected error: %s" % e)

    def unregister_degu(self, mac):
        try:
            resp = self.iot.delete_thing(thingName=mac)
        except Exception as e:
            if e.response['Error']['Code'] == 'ResourceAlreadyExistsException':
                print("User already exists")
            else:
                print("Unexpected error: %s" % e)

    def list_degu(self):
        things = self.iot.list_things(thingTypeName='Degu',  maxResults=2)
        for item in things['things']:
            print(item['thingName'])
        while 'nextToken' in things:
            t = things['nextToken']
            things = self.iot.list_things(thingTypeName='Degu',  maxResults=2, nextToken=t)
            for item in things['things']:
                print(item['thingName'])

    def commission_enable(self):
        proc = subprocess.Popen(['wpanctl','commissioner', 'start'],
                                stdout=subprocess.PIPE)
        while True:
            line = proc.stdout.readline()
            if line != b'':
                print(line)
            else:
                break

    def commission_disable(self):
        proc = subprocess.Popen(['wpanctl','commissioner', 'stop'],
                                stdout=subprocess.PIPE)
        while True:
            line = proc.stdout.readline()
            if line != b'':
                print(line)
            else:
                break

    def commission_joiner(self, eui, pskd):
        proc = subprocess.Popen(['wpanctl','commissioner', 'joiner-add',
                                 eui, '2000', pskd],stdout=subprocess.PIPE)
        while True:
            line = proc.stdout.readline()
            if line != b'':
                print(line)
            else:
                break

    def create_keys_and_certificate(self):
        asset = self.iot.create_keys_and_certificate(setAsActive=True)
        self.id = asset["certificateId"]
        self.arn = asset["certificateArn"]
        return asset

    def attach_thing(self, thingname):
        self.iot.attach_thing_principal(thingName=thingname,
                                        principal=self.arn)

    def get_list_thing_principals(self, mac):
        result = 0
        cert_id = ""
        try:
            list = self.iot.list_thing_principals(thingName=mac)
            result = len(list["principals"])
            if result == 1:
                cert_id_list = list["principals"][0].split("cert/")
                cert_id = cert_id_list[1]
            else:
                pass
        except Exception as e:
            print(e)
        return result, cert_id

    def describe_certificate(self, cert_id):
        status = ""
        try:
            result = self.iot.describe_certificate(certificateId=cert_id)
            status = result['certificateDescription']['status']
            print(status)
        except Exception as e:
            print(e)
        return True, status

    def validate_certificate(self, mac):
        result, cert_id = self.get_list_thing_principals(mac)
        if result == 0:
            print("None")
            status = "None"
        elif result == 1:
            result, status = self.describe_certificate(cert_id)
        else:
            print("Error")
            status = "Error"
        return True, status

    def reattach_certificate(self, mac):
        os.makedirs(ASSET_DIR + mac, exist_ok=True)
        self.create_attach_assets_degu(mac, ASSET_DIR + mac)
        self.attach_thing(mac)
        return True

def gen_static_gw_address():
    out = subprocess.getoutput("wpanctl form -p 0xABCD degu")

    out = subprocess.getoutput(
        """ip -6 addr show dev wpan0 scope global |
        grep inet6 | head -1 |awk '{print $2}'""")

    one  = ipaddress.IPv6Address('::1')
    a = ipaddress.ip_interface(out)
    gw_address = str(ipaddress.IPv6Address(int(a.ip) &
                                           int(a.netmask) | int(one))) + '/64'
    out = subprocess.getoutput("ip -6 addr add "
                               + gw_address +
                               " dev wpan0")

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--register-degu", metavar="MAC", help="register degu by MAC")
    parser.add_argument("--register-gw", action='store_true')
    parser.add_argument("--validate-cert", metavar="MAC", help="validate certificate")
    parser.add_argument("--reattach-cert", metavar="MAC", help="reattach certificate")
    # test
    parser.add_argument("--catest", metavar="MAC", help="reattach certificate")
    args = parser.parse_args()

    if args.register_degu:
        mgr = DeguMgr()
        mgr.register_degu(args.register_degu)
        mgr.commission_disable()
        mgr.commission_enable()
        mgr.commission_joiner(args.register_degu, 'DEGUPSK')

    if args.register_gw:
        gen_static_gw_address()
        mgr = DeguMgr()
        gw_mac = subprocess.getoutput(
            """ip link show dev eth0 | grep link | awk '{print toupper($2)}' | tr -d ':'""")
        mgr.register_gw(gw_mac)

    if args.validate_cert:
        mgr = DeguMgr()
        mgr.validate_certificate(args.validate_cert)

    if args.reattach_cert:
        mgr = DeguMgr()
        mgr.reattach_certificate(args.reattach_cert)

    if args.catest:
        mgr = DeguMgr()
        mgr.register_gw(args.catest)

    exit(0)
